"""Actor model execution framework (see https://en.wikipedia.org/wiki/Actor_model).

For more details, see Thespian.md.
"""


class InvalidActorAddress(Exception):
    '''Exception thrown if an ActorAddress is not valid (i.e. synthesized
    or altered) or could not be generated.'''
    def __init__(self, addr, desc, *args, **kw):
        self.actorAddress = addr
        Exception.__init__(self, str(addr) + ' is ' + desc, *args, **kw)


class ActorSystemStartupFailure(Exception):
    "Exception thrown if ActorSystem cannot startup"
    pass


class NoCompatibleSystemForActor(Exception):
    "Thrown if Actor cannot be started because no ActorSystem matches the needed capabilities"
    def __init__(self, actorClass, msg, *args, **kw):
        Exception.__init__(self, msg + " for Actor {0}".format(str(actorClass)), *args, **kw)



class ActorAddress:
    '''Used to reference a specific Actor (or Troupe).  The Actor could be
    local or remote, in another thread, or another process.  The actor
    is not even guaranteed to be alive anymore.  Regardless, this is
    the identifier by which messages can be directed to an Actor.
    '''
    def __init__(self, aaddr):
        '''Can only be constructed by the ActorSystem(); cannot be constructed
        or synthesized externally.

        The aaddr is meant to be an opaque reference generated
        internally by the ActorSystem implementation and used by it
        for routing purposes.  The user may request a string-version
        of the aaddr, but this should be used for informative or
        display purposes only.

        '''
        self._aaddr = aaddr

    @property
    def addressDetails(self): return self._aaddr

    @property
    def actorAddressString(self):
        if isinstance(self._aaddr, type("")):
            return self._aaddr
        return str(self._aaddr)

    def __str__(self): return 'ActorAddr-%s'%str(self._aaddr)

    def __eq__(self, o):
        if hasattr(self, 'eqOverride'):
            return self.eqOverride(o)
        try:
            return self.actorAddressString == o.actorAddressString
        except:
            return False
    def __ne__(self, o): return not self.__eq__(o)
    def __hash__(self):  return hash(self.actorAddressString)


class Actor(object):
    '''This is the Actor encapsulation itself.  The receiveMessage()
       method should be overridden by the subclass to provide the
       message handling functionality.
    '''

    def __init__(self, **kw):
        """Called to initialize the Actor.

           The kw keyword argument may contain an optional
           childActors.  If this Actor is being instantiated to
           replace a previous Actor instance (i.e. the parent Actor
           responded to the ChildActorExited with a request to replace
           the child) then this argument is a non-empty list of the
           ActorAddress values of the children of the original Actor.

           Override this initialization method as needed in defined Actors.

           N.B.  Currently the Actor is not yet fully realized in the
           ActorSystem when __init__ is invoked.  This means that the
           Actor __init__ cannot invoke any ActorSystem-related
           operations (no .send(), .handleDeadLetters(),
           .notifyOnSystemRegistrationChanges(), etc.)

        """
        pass

    def __str__(self):
        return '{A:' + self.__class__.__name__ + ' @ ' + str(self.myAddress) + '}'

    def receiveMessage(self, msg, sender):
        '''Main entry point handling a request received by this Actor.  Runs
           without interruption and may access locals to this Actor
           (only) without concern that these locals will be modified
           externally.
        '''
        noPyLintWarnings = msg, sender
        assert False, \
            'default Actor.receiveMessage for "%s" must be overridden to handle messages'%self

    @property
    def myAddress(self):
        "Returns the ActorAddress of this Actor itself."
        return self._myRef.address

    def createActor(self, actorClass, targetActorRequirements=None, globalName=None):
        """Initiates creation of a new child Actor of the specified Class.  Returns the ActorAddress for that child
           Actor.

           If the optional globalName parameter is specified, the
           ActorSystem will first check for a registered Actor under
           that name and return that Actor's address if one is
           registered; in this case all other arguments to createActor
           are ignored (and the registered Actor's parent is not
           necessarily the current Actor).  If no Actor is registered
           under that name, the requested Actor is created and
           registered under that name.
        """
        return self._myRef.createActor(actorClass, targetActorRequirements, globalName)

    def send(self, targetAddr, msg):
        """Sends a message to another Actor (specified via ActorAddress) from
           this Actor.  The msg must be pickle-able."""
        if not isinstance(targetAddr, ActorAddress):
            raise ValueError('Target address not a valid ActorAddress')
        self._myRef.actor_send(targetAddr, msg)

    def handleDeadLetters(self, startHandling=True):
        """Registers this Actor with the ActorSystem as a recipient of
           DeadLetters.  If multiple Actors register, the DeadLetter
           is passed to each actor.  If the optional argument is
           False, or if this Actor exits, then it is automatically
           removed from Dead Letter handling. """
        self._myRef.handleDeadLetters(self.myAddress, startHandling)

    def notifyOnSystemRegistrationChanges(self, startHandling=True):
        """Registers this Actor with the ActorSystem as a recipient of 
           ActorSystemConventionUpdate messages."""
        self._myRef.notifyOnSystemRegistrationChanges(self.myAddress, startHandling)

    def logger(self, name=None):
        return self._myRef.logger(name)


class ActorSystemConventionUpdate(object):
    """This message is delivered to actors that have registered for
       ActorSystem Convention updates via their
       notifyOnSystemRegistrationChanges calls.
    """
    def __init__(self, adminAddress, capabilities = None, added = True):
        self._remoteAdminAddress = adminAddress
        self._capabilities       = capabilities
        self._added              = added # if False, remote system actively de-registered


class ActorExitRequest(object):
    """This message should be sent to an Actor to request that it be
       shutdown.  After processing this message, the ActorSystem will
       terminate this Actor instance.  An Actor can stop itself by
       sending this message to itself.
    """
    def __init__(self, permanent=False):
        self._permanent = permanent
    
    def __str__(self): return 'ActorExitRequest'
    
    def isPermanent(self): return self._permanent


class ChildActorExited(object):
    """This is a message sent to any parent Actor (sender is the
    ActorSystem) when one of its child Actors exits.

    """
    def __init__(self, exitedChildAddress, permanent=False):
        self._childAddr = exitedChildAddress
        self._permanent = permanent

    def __str__(self):
        return ('Permanent' if self._permanent else '') + 'ChildActorExited:' + str(self._childAddr)

    @property
    def childAddress(self):
        "Used by the parent Actor to get the address of the exited child"
        return self._childAddr
    
    def isPermanent(self): return self._permanent

    def replaceActor(self, replacementClass):
        """The parent Actor can call this if the exited child should be
           replaced under the same ActorAddress.  Note that the
           replacement is not guaranteed to succeed, and that it might
           result in another ChildActorExited (which may be suppressed
           to prevent an infinite loop).

           The replacementClass may be a string or a class; the latter
           is recommended in case the Actor will be created in a
           remote ActorSystem because the former will likely fail
           import requirements.

           The replacementClass does not allow specification of
           targetActorRequirements (as is possible for the
           createActor); the replacementClass is instantiated in the
           same ActorSystem as the original Actor that exited.  If
           more adjustments should be made, the parent can choose to
           not call replaceActor() and to instantiate a new actor with
           different settings instead (albeit under a different
           ActorAddress).

        """
        # The ActorSystem will override this with actual functionality
        raise NotImplementedError()


class PoisonMessage(object):
    """Message wrapper used to return a message to the sender that has
       caused multiple failures in a recipient Actor."""
    def __init__(self, poison):
        self._poison = poison
    @property
    def poisonMessage(self):
        "Returns the actual message that poisoned the target Actor."
        return self._poison


class ActorSystem(object):
    """Defines the Actor System and external interface operations.

    When initializing, the systemBase can be specified to indicate the
    underlying environment in which the ActorSystem should operate.
    The systemBase can be a string indicating a known base system
    type, or it can be a SystemBase object created by the user.  The
    default SystemBase is the previously specified systemBase, or the
    procSystemBase.ActorSystemBase if this is the first instantiation of
    the ActorSystem.

    Although multiple ActorSystems may be instantiated, it is intended
    that there is only one for the entire application, and that the
    base is specified on the first call only, with subsequent calls
    passing the same base or no base specification (in which case the
    ActorSystem instances will act as a global singleton).  It is
    possible to instantiate different ActorSystems, but the first one
    acts as the Singleton and the others are separate (and may cause
    execution conflicts).

    When injecting messages into the ActorSystem from the outside
    (i.e. not from Actors) the .tell() and .ask() methods should be
    used.  The .tell() method will send a message to the specified
    Actor.  The .ask() method will send a message and then suspend the
    current thread to await a response.

    New Actors can be created from the outside by calling the
    createActor() method; Actors wishing to create new Actors should
    use the .createActor() method on the Actor object.

    A global actor name may be specified as an optional parameter on
    the createActor() method.  This can be used to assign a specific
    name to a specific Actor across the entire ActorSystem.  If there
    is already an Actor registered with this global name, no new actor
    is created and the address of the existing actor is returned.
    There is no other regulation of this namespace and it is
    recommended that it be used sparingly and only for top-level
    Actors that need to be singletons.  Note specifying a globalName
    of an Actor that already exists causes all other createActor
    parameters to be ignored.

    """

    def __init__(self, systemBase=None, capabilities=None, logDefs=None):
        self.systemAddress = ActorAddress('/ActorSys')
        self.capabilities = capabilities or dict()
        if 'logging' in self.capabilities:
            import logging
            logging('Thespian').warning('logging specification moved from capabilities to an explicit argument.')
        if systemBase is None:
            systemBase = getattr(self.__class__, 'systemBase', None)
            if systemBase is None:
                import thespian.system.simpleSystemBase
                systemBase = thespian.system.simpleSystemBase.ActorSystemBase(self, logDefs)
        elif isinstance(systemBase, str):
            import importlib
            # n.b. let standard import exception indicate a missing/unknown systemBase
            module = importlib.import_module('thespian.system.%s'%systemBase)
            systemBase = getattr(module, 'ActorSystemBase')(self, logDefs)
        # else systemBase should be a valid object already
        self._systemBase = systemBase
        if getattr(self.__class__, 'systemBase', None) is None:
            # Set the Singleton systemBase
            self.__class__.systemBase = systemBase

    def shutdown(self):
        "Called to shutdown the ActorSystem itself.  May block until all Actors are shutdown."
        self._systemBase.shutdown()
        delattr(self.__class__, 'systemBase')
        self._systemBase = None

    def createActor(self, actorClass, targetActorRequirements=None, globalName=None):
        'Called to create a "Primary" Actor (a top-level Actor owned by the system itself).'
        return self._systemBase.newPrimaryActor(actorClass, targetActorRequirements, globalName)

    def tell(self, actorAddr, msg):
        "Sends msg to the Actor at the specified address.  No response is expected or awaited."
        if not isinstance(actorAddr, ActorAddress):
            raise ValueError('Actor ask address is not a valid ActorAddress')
        self._systemBase.tell(actorAddr, msg)

    def ask(self, actorAddr, msg, timeout=None):
        """Sends msg to the addressed Actor and waits for a response.  The
        optional timeout argument specifies the maximum amount of time
        to wait in fractional seconds.  Returns None if no response is
        received in the indicated time period.
        """
        if not isinstance(actorAddr, ActorAddress):
            raise ValueError('Actor ask address "%s" is not a valid ActorAddress'%str(actorAddr))
        return self._systemBase.ask(actorAddr, msg, timeout)

    def _handleDeadLetters(self, address, enable):
        self._systemBase._handleDeadLetters(address, enable)

    def systemUpdate(self, updateType, *updateArgs, **updateKWArgs):
        """Back door to allow access to the underlying systemBase
           implementation.  Implementations may differ, so this is not
           generally recommended and is intended for unusual
           circumstances like unit test controls, etc.
        """
        return getattr(self._systemBase, updateType, lambda *a, **kw: None)(*updateArgs, **updateKWArgs)


def requireCapability(cap):
    '''Actor class decorator for requiring a capability.'''
    def go(cls):
        capCheck0 = None
        if hasattr(cls, 'actorSystemCapabilityCheck'): capCheck0 = cls.actorSystemCapabilityCheck
        @staticmethod
        def capCheck1(caps, reqs):
            return caps.get(cap, False) and (capCheck0(caps,reqs) if capCheck0 else True)
        cls.actorSystemCapabilityCheck = capCheck1
        return cls
    return go
