#!/usr/bin/env python

#"This is an administrative tool for the Thespian system."

import cmd
import socket
import traceback
import logging
import datetime
from thespian.actors import *
from thespian.system import (socketAddressMgmt, Thespian_StatusReq,
                             Thespian_SystemStatus, Thespian_ActorStatus)


class ThespianShell(cmd.Cmd):
    intro = "Thespian Actor shell.  Type help or '?' to list commands.'\n"
    prompt = 'thespian> '

    def __init__(self, *args, **kw):
        cmd.Cmd.__init__(self, *args, **kw)
        #super(ThespianShell, self).__init__(*args, **kw)
        self.system = None
        self.knownActorAddresses = []   # val=ActorAddress

    def emptyline(self): pass  # do nothing

    def do_exit(self, arg):
        'Exit the Thespian shell, shutting down any ActorSystem started locally.'
        print ('Exiting.')
        if self.system:
            print ('Shutting down ActorSystem')
            self.system.shutdown()
            del self.system
            self.system = None
        return True

    def do_quit(self, arg):
        "Exits the Thespian shell immediately; no shutdown or cleanup of ActorSystems or Actors."
        self.system = None
        return self.do_exit(arg)

    def do_EOF(self, arg):
        "Exits the Thespian shell immediately; no shutdown or cleanup of ActorSystems or Actors."
        self.system = None
        return self.do_exit(arg)

    def parseActorAddress(self, arg):
        args = dict(zip(['ipaddr', 'port'], arg.split()))
        print ('Args is: %s'%args)
        try:
            return ActorAddress( socketAddressMgmt.SocketActorAddress(
                socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP,
                (args.get('ipaddr', '127.0.0.1'),
                 eval(args.get('port', '14334')))) )
        except:
            print ('***ERROR: unable to parse IP address or port specification')
            traceback.print_exc(limit=3)
            return None


    def parseActorNum(self, arg, useDefault=False):
        if arg.strip():
            try:
                split = arg.split(' ')
                anum = eval(split[0])
                rem  = ' '.join(split[1:])
                if anum >= len(self.knownActorAddresses):
                    print ('***ERROR: Specified Actor Address #%d is not known'%anum)
                    return None
                return anum, self.knownActorAddresses[anum], rem
            except:
                print ('***ERROR parsing Actor number from command line.')
                traceback.print_exc(limit=3)
        else:
            if useDefault:
                if not len(self.knownActorAddresses):
                    self.knownActorAddresses.append(self.parseActorAddress('127.0.0.1 14334'))
                return 0, self.knownActorAddresses[0], ''
        return None


    def getOrAddAddress(self, address):
        for N,A in enumerate(self.knownActorAddresses):
            if A == address:
                return N,A
        self.knownActorAddresses.append(address)
        return len(self.knownActorAddresses) - 1, address

    def do_start(self, arg):
        """Starts an ActorSystem.  The first optional argument is the
           SystemBase.  The remainder of the line (if any) is parsed
           as the capabilities dictionary to pass to the
           ActorSystem.
        """
        if self.system:
            print ('Shutting down previous ActorSystem')
            self.system.shutdown()
            del self.system
            self.system = None
        if arg:
            base = arg.split()[0]
            capspec = ' '.join(arg.split()[1:])
            caps = eval(capspec) if capspec else {}
            print ('Starting %s ActorSystem\nCapabilities: %s'%(base, str(caps)))
            try:
                self.system = ActorSystem(base, caps)
                print ('Started %s ActorSystem'%base)
            except ImportError:
                print ('***ERROR starting ActorSystem with specified Base: %s'%base)
        else:
            self.system = ActorSystem()
            print ('Started default ActorSystem')

    def do_stop(self, arg):
        """Stops any currently-running ActorSystem."""
        ActorSystem().shutdown()
        if self.system:
            print ('Erasing previous ActorSystem')
            del self.system
            self.system = None

    def do_info(self, arg):
        "Gets information about the local environment."
        print ('Thespian local information:')
        if self.system:
            print ('Local system: %s'%str(self.system))
        if hasattr(ActorSystem, 'systemBase'):
            print ('Singleton ActorSystem base: %s'%str(ActorSystem.systemBase.__class__.__name__))
        else:
            print ('No ActorSystem seems to be running')
        for N,A in enumerate(self.knownActorAddresses):
            print ('Actor Address:  %s [#%d]'%(str(A), N))

    def do_python(self, arg):
        "Runs a python command"
        print (eval(arg))


    def do_status(self, arg):
        """Get the status of an Actor or ActorSystem.  The optional argument is the address ID# of the ActorSystem or Actor, defaulting to address ID0, which defaults to (IPv4 127.0.0.1:14334) if not set."""
        actorAddrAndMsg = self.parseActorNum(arg, True)
        if actorAddrAndMsg:
            anum, adminAddr, msg = actorAddrAndMsg
            #print ('Connecting to ActorSystem Admin @ %s (#%d)'%(str(adminAddr), anum))
            response = socketAddressMgmt.askToAddress(adminAddr, Thespian_StatusReq(), 10)
            if response is None:
                print ('***ERROR: no response from %s [#%d]'%(str(adminAddr), anum))
            elif isinstance(response, Thespian_SystemStatus):
                print ('Status of ActorSystem @ %s [#%d]:'%(str(adminAddr), anum))
                print ('  Capabilities[%d]:'%len(response.capabilities))
                for k in response.capabilities:
                    print('    %20s: %s'%(k, response.capabilities[k]))
                if response.conventionLeader:
                    N,A = self.getOrAddAddress(response.conventionLeader)
                    print ('  Convention Leader: %s [#%d]'%(str(A), N))
                    if response.conventionRegisterTime:
                        print ('  Last registered @ %s  (%s ago)'%(
                            str(response.conventionRegisterTime),
                            str(datetime.datetime.now() - response.conventionRegisterTime)))
                elif response.conventionAttendees:
                    print ('  Appears to be the Convention Leader')
                print ('  Convention Attendees [%d]:'%len(response.conventionAttendees))
                for each in response.conventionAttendees:
                    N,A = self.getOrAddAddress(each.remoteAddress)
                    print ('    @ %s [#%d]: %s'%(str(A), N, str(each)))
                print ('  Primary Actors [%d]:'%len(response.childActors))
                for A in response.childActors:
                    N,A = self.getOrAddAddress(A)
                    print ('    @ %s [#%d]'%(str(A), N))
                print ('  Pending Messages [%d]:'%len(response.pendingMessages))
                for F,T,M in response.pendingMessages:
                    NF,AF = self.getOrAddAddress(F)
                    NT,AT = self.getOrAddAddress(T)
                    print ('    %s [#%d] --> %s [#%d]:  %s'%(str(AF), NF, str(AT), NT, M))
                print ('  DeadLetter Addresses [%d]:'%len(response.deadLetterAddresses))
                for A in response.deadLetterAddresses:
                    N,A = self.getOrAddAddress(A)
                    print ('    %s [#%d]'%(str(A), N))
            elif isinstance(response, Thespian_ActorStatus):
                print ('Status of %s Actor @ %s [#%d]:'%(response.actorClass,
                                                         str(adminAddr), anum))
                N,A = self.getOrAddAddress(response.adminAddress)
                print ('  Administrator: %s [#%d]'%(str(A), N))
                print ('  Child Actors [%d]:'%len(response.childActors))
                for A,P in response.childActors:
                    N,A = self.getOrAddAddress(A)
                    print ('    @ %s [#%d], pid %d'%(str(A), N, P))
                print ('  Pending Messages [%d]:'%len(response.pendingMessages))
                for T,M in response.pendingMessages:
                    NT,AT = self.getOrAddAddress(T)
                    print ('    --> %s [#%d]:  %s'%(str(AT), NT, M))
            else:
                print ('Status Query Response to %s [#%d]: %s'%(str(adminAddr),
                                                                anum, str(response)))


    def do_address(self, arg):
        "Adds the IP address and port as a known ActorAddress that can be used.  The first optional argument is the IPv4 address to specify (defaulting to 127.0.0.1) and the second optional argument is the socket (defaulting to 14334).  Alternatively, the single argument 'localAdmin' can be specified, which uses port 14334 on the IPv4 address corresponding to the current system."
        if arg.strip() == 'localAdmin':
            res = socket.getaddrinfo(socket.gethostname(), 14334,
                                     socket.AF_INET, socket.SOCK_STREAM, 0, 0)
            if res:
                actorAddr = ActorAddress(
                    socketAddressMgmt.SocketActorAddress( res[0][0], res[0][1], res[0][2], res[0][4]) )
            else:
                print ('***ERROR: no getaddrinfo response data')
                actorAddr = None
        else:
            actorAddr = self.parseActorAddress(arg)
        if actorAddr:
            N,A = self.getOrAddAddress(actorAddr)
            print ('Actor Address %d:  %s'%(N, str(A)))
        else:
            print ('Not able to determine a valid Actor Address from command-line arguments.')


    def do_create_testActor(self, arg):
        "Creates a Test Actor"
        na = (self.system or ActorSystem()).createActor(TestActor)
        N,A = self.getOrAddAddress(na)
        print ('Created new TestActor %d @ %s'%(N, str(A)))


    def do_tell(self, arg):
        """Sends a message to the identified Actor; does not wait for a response.  The first argument is required and is the Actor number (see create_testActor) and the rest of the line is the message string to send to the Actor."""
        actorAddrAndMsg = self.parseActorNum(arg)
        if actorAddrAndMsg:
            anum, addr, msg = actorAddrAndMsg
            try:
                (self.system or ActorSystem()).tell(addr, msg)
            except:
                print ('***ERROR telling Actor #%d (%s)'%(anum, str(addr)))
                traceback.print_exc(limit=3)


    def do_ask(self, arg):
        """Sends a message to the identified Actor and waits for a response (for up to 10 seconds).  The first argument is required and is the Actor number (see create_testActor) and the rest of the line is the message string to send to the Actor."""
        actorAddrAndMsg = self.parseActorNum(arg)
        if actorAddrAndMsg:
            anum, addr, msg = actorAddrAndMsg
            try:
                r = (self.system or ActorSystem()).ask(addr, msg, 10)
                print ('Actor #%d (%s) response: %s'%(anum, addr, r or '<None... timed out>'))
            except:
                print ('***ERROR asking Actor #%d (%s)'%(anum, str(addr)))
                traceback.print_exc(limit=3)


    def do_kill(self, arg):
        "Kills the specified Actor by sending it an ActorExitRequest. The required argument is the Actor number."
        actorAddrAndMsg = self.parseActorNum(arg)
        if actorAddrAndMsg:
            anum, addr, msg = actorAddrAndMsg
            r = (self.system or ActorSystem()).tell(addr, ActorExitRequest())
            print ('Actor #%d (%s) exit request sent.'%(anum, addr))
                    

class TestActor(Actor):
    def receiveMessage(self, msg, sender):
        logger = logging.getLogger('Thespian.Actor')
        logger.debug('TestActor @ %s got message "%s" from %s',
                     str(self.myAddress), str(msg), str(sender))
        if type(msg) == type(''):
            if msg == 'create':
                self.send(sender, self.createActor(TestActor))
            else:
                self.send(sender, 'TestActor @ %s got: %s'%(str(self.myAddress), str(msg)))


        
if __name__ == "__main__":
    ThespianShell().cmdloop()
