'''The SimpleSystemBase is not a true concurrency environment, but it
 runs in the context of the current thread and simply queues actor
 sends to call each actor's handler in sequence.  This sytem can be
 used for simple actor environments where no parallelism is achieved,
 but any actor doing long-running or suspend (e.g. i/o) activities
 will pause or halt the entire system.

  * Synchronous message delivery

  * Local system only

  * All execution is in the current thread (actor system only runs
    when .tell() or .ask() is called).

  * createActor() always creates the actor instance immediately before returning.

'''

import logging, string, types, functools
from thespian.actors import *
try:
    from logging.config import dictConfig
except ImportError:
    # Old python that doesn't contain this...
    from thespian.system.dictconfig import dictConfig


class ActorRef:
    "Used internally to store the actual actor instance and associated information."
    def __init__(self, actorSystem, parentAddr, addr, inst):
        self._system = actorSystem
        self._oldm = parentAddr
        self._addr = addr
        self._inst = inst  # briefly None until set
        self._yung = []   # actorAddress of children
        # The number of current children is len(self._yung); the
        # childCounter keeps track of the total number of children
        # this Actor has had in its ENTIRE lifetime (i.e. it is not
        # decremented when children are removed from self._yung.  The
        # childCounter is used when generating the unique ActorAddress
        # for a new child.
        self._childCounter = 0
    @property
    def parent(self): return self._oldm
    @property
    def address(self): return self._addr
    @property
    def instance(self): return self._inst
    @instance.setter
    def instance(self, newInstance): self._inst = newInstance
    @property
    def childCount(self): return self._childCounter

    def addChild(self, childAddr):
        self._yung.append(childAddr)
        self._childCounter = self._childCounter + 1
    def removeChild(self, childAddr):
        self._yung = [c for c in self._yung if c != childAddr]
    def listChildren(self):
        return self._yung

    def shutdown(self):
        for child in self._yung:
            self._system._systemBase.actor_send(
                self._system._systemBase.actorRegistry[self._system.systemAddress.actorAddressString].address,
                child, ActorExitRequest())

    # Functionality vectoring for the Actor this represents.
    def createActor(self, actorClass, targetActorRequirements, globalName):
        return self._system._systemBase.newActor(self._addr, actorClass, self._system,
                                                 targetActorRequirements, globalName)

    def actor_send(self, toActorAddr, msg):
        self._system._systemBase.actor_send(self._addr, toActorAddr, msg)

    def handleDeadLetters(self, address, enable):
        self._system._handleDeadLetters(address, enable)

    def notifyOnSystemRegistrationChanges(self, address, enable):
        pass # ignored: simple systems don't have registration

    def logger(self, name=None):
        return logging.LoggerAdapter(logging.getLogger(name),
                                     {'actorAddress': self._addr})


# ----------------------------------------------------------------------

_nameValid    = string.ascii_letters   # what characters are valid in an ActorAddress
_nameValidLen = len(_nameValid)    # precompute the size for convenience

def _namegen(v):
    if v == 0: return 'a'
    x,y = divmod(v, _nameValidLen)
    return (_namegen(x) if x else '') + _nameValid[y]

def _newAddress(prefix, childCount):
    return ActorAddress(prefix + '~' + _namegen(childCount))

def _newChildAddress(parentRef):
    "Returns a new candidate ActorAddress for a child Actor about to be created."
    # Note that this address is not fixed/reserved until the child
    # is actually created (thereby incrementing the parentRef
    # childCount), so calling this multiple times without creating
    # the child may return the same value."
    return _newAddress(parentRef.address.actorAddressString, parentRef.childCount)


# ----------------------------------------------------------------------

class PendingSend:
    "used internally for marshalling pending send operations"
    def __init__(self, sender, msg, toActor):
        self.sender = sender
        self.toActor = toActor
        self.msg = msg
        self.attempts = 0


class BadActor(Actor):   # useable as a "null" Actor which does nothing.
    name = 'BadActor'
    def receiveMessage(self, msg, sender):
        logging.getLogger('Thespian').debug('BadActor discarding message')
        pass  #   Throws away all messages


class External(Actor):
    "Proxy for a requester outside the system; any response will be discarded."
    def receiveMessage(self, msg, sender):
        pass


class ExternalWaiter(Actor):
    '''Temporary actor to hold an (eventual) response for some thread
       external to the ActorSystem to check for and retrieve later.'''
    def __init__(self, *args, **kw):
        super(ExternalWaiter, self).__init__(*args, **kw)
        self.response = None
        self.gotResponse = False
    def receiveMessage(self, msg, sender):
        self.response = msg
        self.gotResponse = True


def actor_base_receive(actorInst, msg, sender):
    logging.getLogger('Thespian').debug(
        'Actor "%s" got message "%s" from "%s"'%(str(actorInst), str(msg), str(sender)))
    actorInst.receiveMessage(msg, sender)


class actorLogFilter(logging.Filter):
    def filter(self, logrecord): return 'actorAddress' in logrecord.__dict__
class notActorLogFilter(logging.Filter):
    def filter(self, logrecord): return 'actorAddress' not in logrecord.__dict__

import sys

defaultLoggingConfig = {
    'version': 1,
    'formatters': {
        'defaultFmt': {
            'format': '%(asctime)s %(levelname)-7s =>  %(message)s  [%(filename)s:%(lineno)s]',
        },
        'actorFmt': {
            'format': '%(asctime)s %(levelname)-7s %(actorAddress)s =>  %(message)s  [%(filename)s:%(lineno)s]',
        },
    },
    'filters': {
        'isActorLog': {'()': actorLogFilter},
        'notActorLog': {'()': notActorLogFilter},
    },
    'handlers': {
        'actorLogHandler': { 'class': 'logging.StreamHandler',
                             'level': 'WARNING',
                             'stream': sys.stderr,
                             'formatter': 'actorFmt',
                             'filters': [ 'isActorLog' ],
                         },
        'regLogHandler': { 'class': 'logging.StreamHandler',
                           'level': 'WARNING',
                           'stream': sys.stderr,
                           'formatter': 'defaultFmt',
                           'filters': [ 'notActorLog' ],
                         },
    },
    'root': { 'handlers': ['actorLogHandler',
                           'regLogHandler',
                       ],
          },
    'disable_existing_loggers': True,
}


class ActorSystemBase:

    def __init__(self, system, logDefs = None):
        self.system = system
        self._pendingSends = []
        if logDefs is not None:
            dictConfig(logDefs or defaultLoggingConfig)
        asys = self._newRefAndActor(system, system.systemAddress,
                                    system.systemAddress,
                                    External)
        extreq = self._newRefAndActor(system, system.systemAddress,
                                      ActorAddress('System:ExternalRequester'),
                                      External)
        badActor = self._newRefAndActor(system, system.systemAddress,
                                        ActorAddress('System:BadActor'), BadActor)
        self.actorRegistry = {  # key=ActorAddress string, value=ActorRef
            system.systemAddress.actorAddressString: asys,
            'System:ExternalRequester': extreq,
            'System:BadActor': badActor,
        }
        self._primaryActors = []
        self._primaryCount  = 0
        self._globalNames = {}
        self.procLimit = 0

    def shutdown(self): pass

    def _runSends(self):
        pausedSends = []
        numsends = 0
        while self._pendingSends:
            numsends += 1
            if self.procLimit and numsends > self.procLimit:
                raise RuntimeError('Too many sends')
            ps = self._pendingSends.pop(0)
            if ps.attempts > 4:
                continue  # discard message if PoisonMessage deliveries are also failing
            elif ps.attempts > 2:
                rcvr, sndr, msg = ps.sender.actorAddressString, ps.toActor, PoisonMessage(ps.msg)
            else:
                rcvr, sndr, msg = ps.toActor.actorAddressString, ps.sender, ps.msg
            tgt = self.actorRegistry.get(rcvr, None) or \
                  self.actorRegistry.get('DeadLetterBox', None)
            if tgt:
                if tgt.instance:
                    if isinstance(ps.msg, ChildActorExited):
                        ps.msg.replaceActor = functools.partial(self._replaceActor, ps.msg)
                    killActor = isinstance(ps.msg, ActorExitRequest)
                    try:
                        # This if is to avoid sending PoisonMessage(ChildActorExited) back to child
                        if not isinstance(ps.msg, ChildActorExited) or ps.msg == msg:
                            tgt.instance._receive(msg, sndr)
                        if isinstance(ps.msg, ChildActorExited):
                            ps2 = pausedSends
                            pausedSends = []
                            for each in ps2:
                                if each.toActor == ps.msg.childAddress:
                                    self._pendingSends.append(each)
                                else:
                                    pausedSends.append(each)
                    except Exception as ex:
                        logging.getLogger('Thespian').warning(
                            'Failure of Actor %s during message processing', tgt.address,
                            exc_info = True)
                        ps.attempts += 1
                        self._pendingSends.append(ps)
                        killActor = True
                    if killActor:
                        self.actorRegistry[ps.toActor.actorAddressString].instance = None
                        for gn in self._globalNames:
                            if self._globalNames[gn] == ps.toActor:
                                del self._globalNames[gn]
                                break
                        tgt._system._systemBase.actor_send(
                            self.actorRegistry[self.system.systemAddress.actorAddressString].address,
                            tgt.parent,
                            ChildActorExited(ps.toActor,
                                             permanent=isinstance(ps.msg, ActorExitRequest)))
                else:
                    # dead Actors that have not had their
                    # ChildActorExited processed by their parent will
                    # have a None instance... just hold onto their
                    # messages for now.
                    pausedSends.append(ps)
                    continue
            else:
                # Target Actor no longer exists.  Handle internal messages and discard all others
                pass

            if isinstance(ps.msg, ChildActorExited):
                deadAddr = ps.msg.childAddress.actorAddressString
                childref = self.actorRegistry.get(deadAddr, None)
                if childref and not childref.instance:
                    # Not replaced, so complete removal, including children
                    childref.shutdown()
                    self.actorRegistry[deadAddr] = None

        self._pendingSends = pausedSends

    @staticmethod
    def _newRefAndActor(actorSystem, parentAddr, actorAddr, actorClass):
        try:
            try:
                actor = actorClass(childActors=None)
            except TypeError as te:
                if "unexpected keyword argument 'childActors'" in str(te):
                    actor = actorClass()
                else:
                    actor = None
        except:
            logging.getLogger('Thespian').warning('Actor total creation error', exc_info=True)
            actor = None
        nar = ActorRef(actorSystem, parentAddr, actorAddr, actor)
        if actor:
            nar.instance._myRef = nar
            nar.instance._receive = types.MethodType(actor_base_receive, nar.instance)
        return nar

    def newPrimaryActor(self, actorClass, targetActorRequirements, globalName):
        "Called internally to create a new Actor instance directly under the ActorSystem."
        if globalName and globalName in self._globalNames:
            return self._globalNames[globalName]
        logger = logging.getLogger('Thespian')
        naa = _newAddress("/A", self._primaryCount)
        self._primaryCount = self._primaryCount + 1
        nar = self._newRefAndActor(self.system, self.system.systemAddress, naa, actorClass)
        if nar.instance:
            if globalName:
                self._globalNames[globalName] = naa
                logger.info('Registered %s as global "%s" Primary Actor',
                            str(naa), globalName)
        if not nar.instance:
            logger.warning('Could not create primary Actor %s @ %s',
                           str(actorClass), str(naa))
            return self.actorRegistry['System:BadActor'].address
        self.actorRegistry[naa.actorAddressString] = nar
        logger.debug('Created primary Actor %s @ %s', str(actorClass), str(naa))
        return naa

    def newActor(self, parentAddr, actorClass, actorSystem, targetActorRequirements, globalName):
        if globalName and globalName in self._globalNames:
            return self._globalNames[globalName]
        pa = self.actorRegistry.get(parentAddr.actorAddressString, None)
        if not pa:
            raise InvalidActorAddress(parentAddr,
                                      'invalid parent Address for new "%s" actor'%str(actorClass))
        naa = _newChildAddress(pa)
        # register *before* creating the child to lock-in the address
        # and because child init failure will result in a
        # ChildActorExited message to this parent.
        pa.addChild(naa)
        # Now create child and add it to the ActorSystem's registry
        self.actorRegistry[naa.actorAddressString] = nar = \
                self._newRefAndActor(actorSystem, parentAddr, naa, actorClass)
        if nar.instance:
            if globalName:
                logging.getLogger('Thespian').info('Registered %s as global "%s" Actor',
                                                   str(naa), globalName)
                self._globalNames[globalName] = naa
        else:
            self.actor_send(self.system.systemAddress, parentAddr, ChildActorExited(naa, permanent=True))
        logging.getLogger('Thespian').debug('Created Actor %s @ %s as child of %s'%(
                                             str(actorClass), str(naa), str(parentAddr)))
        return naa

    def _replaceActor(self, childExitedMsg, replacementClass):
        aref = self.actorRegistry.get(childExitedMsg.childAddress.actorAddressString, None)
        if aref:
            try:
                try:
                    actor = replacementClass(childActors=aref.listChildren())
                except TypeError as te:
                    if "unexpected keyword argument 'childActors'" in str(te):
                        actor = replacementClass()
                    else:
                        raise
                aref.instance = actor
                aref.instance._myRef = aref
                aref.instance._receive = types.MethodType(actor_base_receive, aref.instance)
            except Exception as ex:
                logging.getLogger('Thespian').error((
                    'Replacement of actor @ %s with %s failed;' + \
                    ' no further restarts, address dead-lettered. (Exception %s)')%(
                        childExitedMsg.childAddress,
                        replacementClass, str(ex)))
        else:
            logging.getLogger('Thespian').warning(
                'Replacing actor that no longer exists @ %s... ' + \
                'this is probably an ActorSystem problem, ' + \
                'but no replacement will be made.'%childExitedMsg.childAddress)

    def tell(self, anActor, msg):
        sender = ActorAddress('System:ExternalRequester')
        self._pendingSends.append(PendingSend(sender, msg, anActor))
        self._runSends()

    def ask(self, anActor, msg, timeout):
        awa = self.newActor(self.system.systemAddress, ExternalWaiter, self.system,
                            None, None) # new per ask
        aw = self.actorRegistry[awa.actorAddressString].instance
        self._pendingSends.append(PendingSend(awa, msg, anActor))
        # timeout is ignored because we are executing in the context
        # of the current thread, so all actors will run to completion
        # synchronously (or block on external operations) and timeout
        # cannot be effectively implemented.  At best, runSends could
        # check remaining time between Actor calls and return if the
        # timeout period has been exceeded, but that still wouldn't
        # allow interruption of blocked Actors.
        self._runSends()
        response = aw.response
        if not aw.gotResponse:
            logging.getLogger('Thespian').warning('No response received for ask message to %s'%(
                str(anActor)))
        self._pendingSends.append(PendingSend(ActorAddress('System:ExternalRequester'),
                                              ActorExitRequest(), awa))
        self._runSends()
        return response

    def actor_send(self, fromActor, toActor, msg):
        self._pendingSends.append(PendingSend(fromActor, msg, toActor))

    def _handleDeadLetters(self, address, enable):
        reg = self.actorRegistry.get(address.actorAddressString, None)
        if enable:
            self.actorRegistry['DeadLetterBox'] = reg
        else:
            if reg == self.actorRegistry.get('DeadLetterBox', None):
                self.actorRegistry['DeadLetterBox'] = None

    def setProcessingLimit(self, limit=0):
        self.procLimit = limit
