import unittest
import thespian.test.helpers
from thespian.actors import *
from thespian.test import ActorSystemTestCase



class Parent(Actor):
    def __init__(self, *args, **kw):
        super(Parent, self).__init__(*args, **kw)
        self.son = None      # Always a NonStarter
        self.daughter = None  # Starter on restart
        self.poisonedChild = False
    def receiveMessage(self, msg, sender):
        if msg == 'have a son?':
            if not self.son:
                self.son = self.createActor(NonStarter)
            self.send(sender, self.son if self.son else 'no')
        elif msg == 'have a daughter?':
            if not self.daughter:
                self.daughter = self.createActor(NonStarter)
            self.send(sender, self.daughter if self.daughter else 'no')
        elif msg == 'name?':
            self.send(sender, self.myAddress)
        elif isinstance(msg, Deadly):
            if msg.countdown:
                self.send(self.son or self.daughter, Deadly(msg.countdown - 1))
            else:
                raise ValueError('Deadly value received!')
        elif isinstance(msg, PoisonMessage):
            self.poisonedChild = True
        elif msg == 'poisoned child?':
            self.send(sender, self.poisonedChild)
        elif isinstance(msg, KillReq):
            if msg.countdown:
                self.send(self.son or self.daughter, KillReq(msg.countdown - 1))
            else:
                self.send(self.myAddress, ActorExitRequest())


class RestartParent(Parent):
    "Sons that exit are restarted with NonStarter, all other children are restarted as RestartParent."
    def __init__(self, *args, **kw):
        super(RestartParent, self).__init__(*args, **kw)
    def receiveMessage(self, msg, sender):
        if isinstance(msg, ChildActorExited):
            msg.replaceActor(NonStarter if self.son and msg.childAddress == self.son
                             else RestartParent)
        else:
            super(RestartParent, self).receiveMessage(msg, sender)


class NoRestartParent(Parent):
    def __init__(self, *args, **kw):
        super(NoRestartParent, self).__init__(*args, **kw)
    def receiveMessage(self, msg, sender):
        if isinstance(msg, ChildActorExited):
            pass # ignore this... child not restarted
        else:
            super(NoRestartParent, self).receiveMessage(msg, sender)


class NonStarter(Actor):
    def __init__(self, *args, **kw):
        raise Exception('This actor can never start')
    def receiveMessage(self, msg, sender):
        if not isinstance(msg, ActorExitRequest):
            self.send(sender, msg)

class Deadly(object):
    def __init__(self, v):
        self.countdown = v

class KillReq(object):
    def __init__(self, v):
        self.countdown = v


class TestASimpleSystemActorFailures(ActorSystemTestCase):

    def test01_NonStartingSystemLevelActor(self):
        nonstarter = ActorSystem().createActor(NonStarter)
        # just finish, make sure no exception is thrown.  Primary
        # actors (those owned by the ActorSystem itself) are not
        # restarted on failure, so the actor won't actually be
        # recreated.
        self.assertIsNone(ActorSystem().ask(nonstarter, "anything", 0.3))

    def test02_NonStartingSubActorWithRestarts(self):
        parent = ActorSystem().createActor(RestartParent)
        self.assertEqual(ActorSystem().ask(parent, 'name?'), parent)
        son = ActorSystem().ask(parent, 'have a son?')
        self.assertIsNot(son, None)
        self.assertIsNone(ActorSystem().ask(son, 'name?', 0.3))  # dead-lettered, so no response

        self.assertEqual(ActorSystem().ask(parent, 'name?', 4), parent)
        ActorSystem().tell(parent, ActorExitRequest())
        self.assertEqual(ActorSystem().ask(parent, 'name?', 0.4), None)

    def test03_NonStartingSubActorWithoutRestarts(self):
        parent = ActorSystem().createActor(NoRestartParent)
        self.assertEqual(ActorSystem().ask(parent, 'name?', 4), parent)
        son = ActorSystem().ask(parent, 'have a son?', 4)
        self.assertIsNot(son, None)
        self.assertIsNone(ActorSystem().ask(son, 'name?', 0.3))  # dead-lettered, so no response

    def test04_RestartedSubActorWithRestarts(self):
        parent = ActorSystem().createActor(RestartParent)
        self.assertEqual(ActorSystem().ask(parent, 'name?', 4), parent)
        kid = ActorSystem().ask(parent, 'have a daughter?', 4)
        self.assertIsNot(kid, None)
        self.assertEqual(kid, ActorSystem().ask(kid, 'name?', 4))

        self.assertEqual(ActorSystem().ask(parent, 'name?', 4), parent)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 4), kid)
        ActorSystem().tell(parent, ActorExitRequest())
        # root Actors are not restarted which should cause children to be shutdown.
        self.assertEqual(ActorSystem().ask(parent, 'name?', 0.2), None)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 0.2), None)

    def test05_RestartedSubActorWithoutRestarts(self):
        parent = ActorSystem().createActor(NoRestartParent)
        self.assertEqual(ActorSystem().ask(parent, 'name?', 4), parent)
        kid = ActorSystem().ask(parent, 'have a daughter?', 4)
        self.assertIsNot(kid, None)
        self.assertIsNone(ActorSystem().ask(kid, 'name?', 0.5))  # dead-lettered, so no response

    def test06_ActorStackShutdown(self):
        parent = ActorSystem().createActor(RestartParent)
        kid = ActorSystem().ask(parent, 'have a daughter?', 4)
        grandkid = ActorSystem().ask(kid, 'have a daughter?', 4)
        greatgrandkid = ActorSystem().ask(grandkid, 'have a daughter?', 4)

        self.assertEqual(ActorSystem().ask(parent, 'name?', 4), parent)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 4), kid)
        self.assertEqual(ActorSystem().ask(grandkid, 'name?', 4), grandkid)
        self.assertEqual(ActorSystem().ask(greatgrandkid, 'name?', 4), greatgrandkid)

        ActorSystem().tell(parent, ActorExitRequest())
        #time.sleep(0.2)  # allow actor shutdown requests to propagate

        # False positive (success) if actor system is hung?  Need to
        # check deadletter delivery of these name queries to be sure
        # ActorSystem is fully functional.
        self.assertEqual(ActorSystem().ask(parent, 'name?', 0.5), None)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 0.5), None)
        self.assertEqual(ActorSystem().ask(grandkid, 'name?', 0.5), None)
        self.assertEqual(ActorSystem().ask(greatgrandkid, 'name?', 0.5), None)

    def test07_DeepActorShutdown(self):
        parent = ActorSystem().createActor(RestartParent)
        kid = ActorSystem().ask(parent, 'have a daughter?', 4)
        grandkid = ActorSystem().ask(kid, 'have a daughter?', 4)
        greatgrandkid = ActorSystem().ask(grandkid, 'have a daughter?', 4)

        self.assertEqual(ActorSystem().ask(parent, 'name?', 4), parent)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 4), kid)
        self.assertEqual(ActorSystem().ask(grandkid, 'name?', 4), grandkid)
        self.assertEqual(ActorSystem().ask(greatgrandkid, 'name?', 4), greatgrandkid)

        ActorSystem().tell(greatgrandkid, ActorExitRequest())
        #time.sleep(0.2)  # allow actor shutdown requests to propagate

        self.assertEqual(ActorSystem().ask(parent, 'name?', 4), parent)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 4), kid)
        self.assertEqual(ActorSystem().ask(grandkid, 'name?', 4), grandkid)
        self.assertEqual(ActorSystem().ask(greatgrandkid, 'name?', 4), greatgrandkid)

        ActorSystem().tell(kid, ActorExitRequest())
        #time.sleep(0.2)  # allow actor shutdown requests to propagate

        self.assertEqual(ActorSystem().ask(parent, 'name?', 4), parent)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 4), kid)
        self.assertEqual(ActorSystem().ask(grandkid, 'name?', 4), grandkid)
        self.assertEqual(ActorSystem().ask(greatgrandkid, 'name?', 4), greatgrandkid)

        ActorSystem().tell(parent, ActorExitRequest())
        #time.sleep(0.2)  # allow actor shutdown requests to propagate

        # False positive (success) if actor system is hung?  Need to
        # check deadletter delivery of these name queries to be sure
        # ActorSystem is fully functional.
        self.assertEqual(ActorSystem().ask(parent, 'name?', 0.5), None)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 0.5), None)
        self.assertEqual(ActorSystem().ask(grandkid, 'name?', 0.5), None)
        self.assertEqual(ActorSystem().ask(greatgrandkid, 'name?', 0.5), None)

    def test08_DeepActorInvoluntaryTermination(self):
        parent = ActorSystem().createActor(RestartParent)
        kid = ActorSystem().ask(parent, 'have a daughter?', 4)
        grandkid = ActorSystem().ask(kid, 'have a daughter?', 4)
        greatgrandkid = ActorSystem().ask(grandkid, 'have a daughter?', 4)

        self.assertEqual(ActorSystem().ask(parent, 'name?', 4), parent)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 4), kid)
        self.assertEqual(ActorSystem().ask(grandkid, 'name?', 4), grandkid)
        self.assertEqual(ActorSystem().ask(greatgrandkid, 'name?', 4), greatgrandkid)

        self.assertFalse(ActorSystem().ask(parent, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(kid, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(grandkid, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(greatgrandkid, 'poisoned child?', 4))

        ActorSystem().tell(parent, Deadly(3))  # kills greatgrandkid
        #time.sleep(0.75)

        self.assertEqual(ActorSystem().ask(parent, 'name?', 4), parent)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 4), kid)
        self.assertEqual(ActorSystem().ask(grandkid, 'name?', 4), grandkid)
        self.assertEqual(ActorSystem().ask(greatgrandkid, 'name?', 4), greatgrandkid)

        self.assertFalse(ActorSystem().ask(parent, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(kid, 'poisoned child?', 4))
        self.assertTrue(ActorSystem().ask(grandkid, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(greatgrandkid, 'poisoned child?', 4))

        ActorSystem().tell(parent, Deadly(1))  # kills kid
        #time.sleep(0.28)  # wait for Deadly message effects to propagate

        self.assertEqual(ActorSystem().ask(parent, 'name?', 4), parent)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 4), kid)
        self.assertEqual(ActorSystem().ask(grandkid, 'name?', 4), grandkid)
        self.assertEqual(ActorSystem().ask(greatgrandkid, 'name?', 4), greatgrandkid)

        self.assertTrue(ActorSystem().ask(parent, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(kid, 'poisoned child?', 4))
        self.assertTrue(ActorSystem().ask(grandkid, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(greatgrandkid, 'poisoned child?', 4))

        ActorSystem().tell(parent, Deadly(0))  # kills parent; no restarts for top level
        #time.sleep(0.38)  # wait for Deadly message effects to propagate

        self.assertEqual(ActorSystem().ask(parent, 'name?', 0.25), None)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 0.25), None)
        self.assertEqual(ActorSystem().ask(grandkid, 'name?', 0.25), None)
        self.assertEqual(ActorSystem().ask(greatgrandkid, 'name?', 0.25), None)

    def test09_DeepActorSuicide(self):
        parent = ActorSystem().createActor(RestartParent)
        kid = ActorSystem().ask(parent, 'have a daughter?', 4)
        grandkid = ActorSystem().ask(kid, 'have a daughter?', 4)
        greatgrandkid = ActorSystem().ask(grandkid, 'have a daughter?', 4)

        self.assertEqual(ActorSystem().ask(parent, 'name?', 4), parent)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 4), kid)
        self.assertEqual(ActorSystem().ask(grandkid, 'name?', 4), grandkid)
        self.assertEqual(ActorSystem().ask(greatgrandkid, 'name?', 4), greatgrandkid)

        self.assertFalse(ActorSystem().ask(parent, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(kid, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(grandkid, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(greatgrandkid, 'poisoned child?', 4))

        ActorSystem().tell(parent, KillReq(3))  # kills greatgrandkid

        self.assertEqual(ActorSystem().ask(parent, 'name?', 4), parent)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 4), kid)
        self.assertEqual(ActorSystem().ask(grandkid, 'name?', 4), grandkid)
        self.assertEqual(ActorSystem().ask(greatgrandkid, 'name?', 4), greatgrandkid)

        self.assertFalse(ActorSystem().ask(parent, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(kid, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(grandkid, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(greatgrandkid, 'poisoned child?', 4))

        ActorSystem().tell(parent, KillReq(1))  # kills kid

        self.assertEqual(ActorSystem().ask(parent, 'name?', 4), parent)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 4), kid)
        self.assertEqual(ActorSystem().ask(grandkid, 'name?', 4), grandkid)
        self.assertEqual(ActorSystem().ask(greatgrandkid, 'name?', 4), greatgrandkid)

        self.assertFalse(ActorSystem().ask(parent, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(kid, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(grandkid, 'poisoned child?', 4))
        self.assertFalse(ActorSystem().ask(greatgrandkid, 'poisoned child?', 4))

        ActorSystem().tell(parent, KillReq(0))  # kills parent; no restarts for top level

        ActorSystem().ask(parent, 'name?', 0.2)  # throw-away to allow KillReq to be processed.
        self.assertEqual(ActorSystem().ask(parent, 'name?', 0.31), None)
        self.assertEqual(ActorSystem().ask(kid, 'name?', 0.31), None)
        self.assertEqual(ActorSystem().ask(grandkid, 'name?', 0.31), None)
        self.assertEqual(ActorSystem().ask(greatgrandkid, 'name?', 0.31), None)

#until multiprocess system works on Windows
import os
if os.name!='nt':
    class TestMultiProcessingSystemActorFailures(TestASimpleSystemActorFailures):
        def setUp(self):
            self.setSystemBase('multiprocessSystemBase')
